/*

 Copyright (C) 1990-1994 Mark Adler, Richard B. Wales, Jean-loup Gailly,
 Kai Uwe Rommel and Igor Mandrichenko.
 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as all of the original files are included,
 that it is not sold for profit, and that this copyright notice is retained.

*/

/*
 *  zipsort.c by Mark Adler.
 */

#define UTIL
#include "revision.h"
#include "zip.h"
#include <signal.h>


/* Temporary zip file name and file pointer */
local char *tempzip;
local FILE *tempzf;


/* Local functions */
#ifdef PROTO
   local void handler(int);
   local void license(void);
   local void help(void);
   local int keylt(char *, char *);
   local void heapsort(extent, char far * far *);
   int main(int, char **);
#endif /* PROTO */



void err(c, h)
int c;                  /* error code from the ZE_ class */
char *h;                /* message about how it happened */
/* Issue a message for the error, clean up files and memory, and exit. */
{
  if (PERR(c))
    perror("zipsort error");
  fprintf(stderr, "zipsort error: %s (%s)\n", errors[c-1], h);
  if (tempzf != NULL)
    fclose(tempzf);
  if (tempzip != NULL)
  {
    destroy(tempzip);
    free((voidp *)tempzip);
  }
  if (zipfile != NULL)
    free((voidp *)zipfile);
#ifdef VMS
  exit(0);
#else /* !VMS */
  exit(c);
#endif /* ?VMS */
}


local void handler(s)
int s;                  /* signal number (ignored) */
/* Upon getting a user interrupt, abort cleanly using err(). */
{
#ifndef MSDOS
  putc('\n', stderr);
#endif /* !MSDOS */
  err(ZE_ABORT, "aborting");
  s++;                                  /* keep some compilers happy */
}


void warn(a, b)
char *a, *b;            /* message strings juxtaposed in output */
/* Print a warning message to stderr and return. */
{
  fprintf(stderr, "zipsort warning: %s%s\n", a, b);
}


local void license()
/* Print license information to stdout. */
{
  extent i;             /* counter for copyright array */

  for (i = 0; i < sizeof(copyright)/sizeof(char *); i++) {
    printf(copyright[i], "zipsort");
    putchar('\n');
  }
  for (i = 0; i < sizeof(disclaimer)/sizeof(char *); i++)
    puts(disclaimer[i]);
}


local void help()
/* Print help (along with license info) to stdout. */
{
  extent i;             /* counter for help array */

  /* help array */
  static char *text[] = {
"",
"ZipSort %s (%s)",
"Usage:  zipsort [-b path] zipfile",
"  the action is to sort the entries by name",
"  -b   use \"path\" for the temporary zip file",
"  -h   show this help               -L   show software license",
"",
"Example:",
"     zipsort foo.zip"
  };

  for (i = 0; i < sizeof(copyright)/sizeof(char *); i++) {
    printf(copyright[i], "zipsort");
    putchar('\n');
  }
  for (i = 0; i < sizeof(text)/sizeof(char *); i++)
  {
    printf(text[i], VERSION, REVDATE);
    putchar('\n');
  }
}



int keylt(p, q)
char *p, *q;
{
  return strcmp(((struct zlist far *)p)->zname,
                ((struct zlist far *)q)->zname) < 0;
}


void heapsort(n, a)
extent n;
char far * far *a;
/* algorithm from Numerical Recipes in C, 2nd Ed. */
{
  extent i, r, j, k;
  char far *e;

  if (n < 2)
    return;
  k = n >> 1;                   /* hire supervisors from here down */
  r = n - 1;                    /* retire and promote from here down */
  while (1)
  {
    if (k > 0)                  /* hiring phase */
      e = a[--k];
    else                        /* retirement and promotion phase */
    {
      e = a[r];
      a[r] = a[0];              /* retire top guy to end */
      if (--r == 0)
      {
        a[0] = e;               /* least competent to top! */
        break;
      }
    }
    i = k;
    j = k + k + 1;
    while (j <= r)              /* sift e down to proper level */
    {
      if (j < r && keylt(a[j], a[j + 1]))
        j++;                    /* go to better underling */
      if (keylt(e, a[j]))       /* demote e if underling better */
      {
        a[i] = a[j];
        i = j;
        j++;  j <<= 1;  j--;
      }
      else                      /* else this is e's level */
        break;
    }
    a[i] = e;                   /* put e in his slot */
  }
}


int main(argc, argv)
int argc;               /* number of tokens in command line */
char **argv;            /* command line tokens */
/* Sort the zipfile */
{
  ulg c;                /* start of central directory */
  extent j;             /* counter */
  int k;                /* next argument type */
  char *q;              /* steps through option arguments */
  int r;                /* arg counter, temporary variable */
  ulg s;                /* length of central directory */
  int t;                /* attributes of zip file */
  FILE *x, *y;          /* input and output zip files */
  struct zlist far *z;  /* steps through zfiles linked list */
  struct zlist far * far *i;    /* index pointer list */


  /* If no args, show help */
  if (argc == 1)
  {
    help();
    exit(0);
  }

  init_upper();           /* build case map table */

  /* Go through args */
  zipfile = tempzip = NULL;
  tempzf = NULL;
  signal(SIGINT, handler);
#ifdef SIGTERM              /* AMIGA has no SIGTERM */
  signal(SIGTERM, handler);
#endif
  k = 0;
  for (r = 1; r < argc; r++)
    if (*argv[r] == '-')
      if (argv[r][1])
        for (q = argv[r]+1; *q; q++)
          switch(*q)
          {
            case 'b':   /* Specify path for temporary file */
              if (k)
                err(ZE_PARMS, "use -b before zip file name");
              else
                k = 1;          /* Next non-option is path */
              break;
            case 'h':   /* Show help */
              help();  exit(0);
            case 'l':  case 'L':  /* Show copyright and disclaimer */
              license();  exit(0);
            default:
              err(ZE_PARMS, "unknown option");
          }
      else
        err(ZE_PARMS, "zip file cannot be stdin");
    else
      if (k == 0)
        if (zipfile == NULL)
        {
          if ((zipfile = ziptyp(argv[r])) == NULL)
            err(ZE_MEM, "was processing arguments");
        }
        else
          err(ZE_PARMS, "can only specify one zip file");
      else
      {
        tempath = argv[r];
        k = 0;
      }
  if (zipfile == NULL)
    err(ZE_PARMS, "need to specify zip file");

  /* Read zip file */
  if ((r = readzipfile()) != ZE_OK)
    err(r, zipfile);
  if (zfiles == NULL)
    err(ZE_NAME, zipfile);

  /* Make sure zip file is writeable */
  if ((x = fopen(zipfile, "a")) == NULL)
    err(ZE_CREAT, zipfile);
  fclose(x);
  t = getfileattr(zipfile);

  /* Make index list to entries */
  if ((i = (struct zlist far * far *)farmalloc(zcount * sizeof(z))) == NULL)
    err(ZE_MEM, "was sorting names");
  for (z = zfiles; z != NULL; z = z->nxt)
    *i++ = z;
  i -= zcount;

  /* Sort index list (use heapsort to guarantee good behavior) */
  heapsort(zcount, (char far * far *)i);

  /* Reorder entries based on now sorted list */
  z = zfiles = i[0];
  for (j = 1; j < zcount; j++)
    z = z->nxt = i[j];
  z->nxt = NULL;

  /* Open output zip file for writing */
  if ((tempzf = y = fopen(tempzip = tempname(zipfile), FOPW)) == NULL)
    err(ZE_TEMP, tempzip);

  /* Open input zip file again, copy preamble if any */
  if ((x = fopen(zipfile, FOPR)) == NULL)
    err(ZE_NAME, zipfile);
  if (zipbeg && (r = fcopy(x, y, zipbeg)) != ZE_OK)
    err(r, r == ZE_TEMP ? tempzip : zipfile);

  /* Go through local entries, copying them over as is */
  for (z = zfiles; z != NULL; z = z->nxt)
    if ((r = zipcopy(z, x, y)) != ZE_OK)
      err(r, "was copying an entry");
  fclose(x);

  /* Write central directory and end of central directory */
  if ((c = ftell(y)) == -1L)    /* get start of central */
    err(ZE_TEMP, tempzip);
  for (z = zfiles; z != NULL; z = z->nxt)
    if ((r = putcentral(z, y)) != ZE_OK)
      err(r, tempzip);
  if ((s = ftell(y)) == -1L)    /* get end of central */
    err(ZE_TEMP, tempzip);
  s -= c;                       /* compute length of central */
  if ((r = putend((int)zcount, s, c, zcomlen, zcomment, y)) != ZE_OK)
    err(r, tempzip);
  tempzf = NULL;
  if (fclose(y))
    err(ZE_TEMP, tempzip);
  if ((r = replace(zipfile, tempzip)) != ZE_OK)
  {
    warn("new zip file left as: ", tempzip);
    free((voidp *)tempzip);
    tempzip = NULL;
    err(r, "was replacing the original zip file");
  }
  free((voidp *)tempzip);
  tempzip = NULL;
  setfileattr(zipfile, t);
  free((voidp *)zipfile);
  zipfile = NULL;

  /* Done! */
  exit(ZE_OK);
  return 0; /* avoid warning */
}
